# 키보드 입력을 사용하기 때문에 반드시 콘솔창에서 실행해야 합니다.

from roboid import *
import random

hamster = Hamster()

ROWS = 4 # 말판 행(row)의 개수 (세로 방향 개수)
COLS = 4 # 말판 열(column)의 개수 (가로 방향 개수)

UP = 0
RIGHT = 1
DOWN = 2
LEFT = 3
ACTIONS = [UP, RIGHT, DOWN, LEFT]

class Mover:
    def __init__(self):
        self.reset()
        self.counts = []
        self.move_count = 0
        self.q = self.make_q()
        self.wait_space_key()

    def reset(self): # 로봇의 위치, 방향 초기화 (출발 위치와 방향)
        self.x = 1
        self.y = 1
        self.direction = RIGHT
        hamster.stop()

    def make_q(self): # 로봇의 위치가 0부터가 아닌 1부터 시작하므로 한 줄 더 (말판이 그렇게 되어 있음)
        q = [ [None] * (COLS+1) for _ in range(ROWS+1) ]
        for y in range(1, ROWS+1):
            for x in range(1, COLS+1):
                q[y][x] = [0] * len(ACTIONS)
        return q

    def is_valid_action(self, x, y, action): # (x, y) 위치에서 action을 수행할 수 있는가? True/False
        if action == LEFT: return x > 1
        elif action == RIGHT: return x < COLS
        elif action == UP: return y < ROWS
        else: return y > 1 # DOWN

    def direction_diff(self, direction): # 목표 방향(direction)과 현재 방향의 차이 계산
        return (direction - self.direction) % 4

    def turn_to(self, direction): # 로봇을 direction 방향으로 회전
        turn = self.direction_diff(direction)
        if turn == 1: # 시계 방향 90도
            hamster.board_right()
        elif turn == 2: # 반대 방향 180도
            hamster.board_left()
            hamster.board_left()
        elif turn == 3: # 반시계 방향 90도
            hamster.board_left()
        self.direction = direction

    def move(self, action): # 로봇을 action에 따라 이동
        self.turn_to(action)
        hamster.board_forward()
        if action == LEFT: self.x -= 1
        elif action == RIGHT: self.x += 1
        elif action == UP: self.y += 1
        else: self.y -= 1 # DOWN

    def move_to(self, x, y): # 로봇을 (x, y) 위치로 이동
        if x > self.x:
            for _ in range(x - self.x):
                self.move(RIGHT)
        elif x < self.x:
            for _ in range(self.x - x):
                self.move(LEFT)
        if y > self.y:
            for _ in range(y - self.y):
                self.move(UP)
        elif y < self.y:
            for _ in range(self.y - y):
                self.move(DOWN)

    def move_to_start(self): # 로봇을 출발 위치로 이동
        self.move_to(1, 1)
        self.turn_to(RIGHT)

    def calc_best_action(self, x, y): # (x, y) 위치에서 q 값이 가장 큰 행동은?
        q_max = float('-inf')
        best = []
        for a in ACTIONS:
            if self.is_valid_action(x, y, a) and self.direction_diff(a) != 2: # 180도 반대 방향으로 이동하는 행동은 제외
                q_value = self.q[y][x][a]
                if q_value > q_max:
                    q_max = q_value
                    best = [a]
                elif q_value == q_max: # q 값이 같은 것이 있으면 랜덤으로 선택
                    best.append(a)
        if len(best) == 1:
            return best[0]
        else:
            return random.choice(best)

    def get_max_q(self, x, y): # (x, y) 위치에서 최대 q 값은?
        q_max = float('-inf')
        for a in ACTIONS:
            if self.is_valid_action(x, y, a):
                q_value = self.q[y][x][a]
                if q_value > q_max:
                    q_max = q_value
        return q_max

    def update(self, x, y, action, reward, key): # (x, y) 위치의 행동(action)에 대한 q 값을 보상(reward) 적용하여 갱신
        # (x, y): 이동하기 전의 위치
        # (self.x, self.y) 이동한 후의 위치
        self.q[y][x][action] = reward + 0.9 * self.get_max_q(self.x, self.y)

        if key == 'o':
            self.counts.append(self.move_count)
            self.move_count = 0
            print(self.counts)
            hamster.leds('green')
            hamster.note('c4', 0.5)
            hamster.note('e4', 0.5)
            hamster.note('g4', 0.5)
            hamster.leds('off')
            self.move_to_start()
            self.wait_space_key()
        elif key == 'x':
            hamster.leds('red')
            hamster.beep()
            hamster.beep()
            hamster.leds('off')
            self.move_to_start()
            self.wait_space_key()

    def wait_space_key(self):
        while True:
            key = Keyboard.read()
            if key:
                if key == 'r':
                    self.reset()
                elif key == ' ':
                    break
            wait(20) # 무한 반복은 너무 빨리 하면 통신을 못하므로 잠깐씩 쉬어야 함(10 ~ 20 msec)

    def check_key(self):
        while True:
            key = Keyboard.read()
            if key:
                if key == 'r':
                    self.reset()
                elif key == ' ' or key == 'o' or key == 'x':
                    return key
            wait(20) # 무한 반복은 너무 빨리 하면 통신을 못하므로 잠깐씩 쉬어야 함(10 ~ 20 msec)

    def run_once(self):
        x = self.x # 현재 위치
        y = self.y
        action = self.calc_best_action(x, y) # 현재 위치에서 최적의 행동(action)
        self.move(action) # 최적의 행동에 따라 이동. 현재 위치와 방향 변경됨
        self.move_count += 1
        key = self.check_key() # 키를 누를 때까지 기다림
        
        reward = 0
        if key == 'o': reward = 1
        elif key == 'x': reward = -1
        
        self.update(x, y, action, reward, key) # q값 갱신 및 키에 따른 로봇의 행동(좋아요, 싫어요) 후 출발 위치로 이동

    def run(self):
        while True:
            self.run_once()
            wait(20) # 무한 반복은 너무 빨리 하면 통신을 못하므로 잠깐씩 쉬어야 함(10 ~ 20 msec)

mover = Mover()
mover.run()
